//! Saving and loading chain
use super::*;
use std::{
    sync::Arc,
    path::{
	Path,
    },
    io,
};
use tokio::{
    time::{
	self,
	Duration,
    },
    fs::{
	OpenOptions,
    },
    prelude::*,
};
use futures::{
    future::{
	OptionFuture,
    },
};
use lzzzz::{
    lz4f::{
	self,
	AsyncWriteCompressor,
	PreferencesBuilder,
	AsyncReadDecompressor,
    },
};

const SAVE_INTERVAL: Option<Duration> = Some(Duration::from_secs(2));


pub async fn save_now(state: &State) -> io::Result<()>
{
    let chain = state.chain().read().await;
    use std::ops::Deref;
    let to = &state.config().file;
    save_now_to(chain.deref(),to).await
}

async fn save_now_to(chain: &Chain<String>, to: impl AsRef<Path>) -> io::Result<()>
{
    debug!("Saving chain to {:?}", to.as_ref());
    let mut file = OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .open(to).await?;
    let chain = serde_cbor::to_vec(chain).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))?;
    { 
        let mut file = AsyncWriteCompressor::new(&mut file, PreferencesBuilder::new()
                .compression_level(lz4f::CLEVEL_HIGH).build())?;
        file.write_all(&chain[..]).await?;
        file.flush().await?;
        file.shutdown().await?;
    }
    file.flush().await?;
    file.shutdown().await?;
    Ok(())
}

/// Start the save loop for this chain
pub async fn host(state: State)
{
    let to = &state.config().file;
    let interval = state.config().save_interval();
    while Arc::strong_count(state.when()) > 1 {
	{
	    let chain = state.chain().read().await;
	    use std::ops::Deref;
	    if let Err(e) = save_now_to(chain.deref(), &to).await {
		error!("Failed to save chain: {}", e);
	    } else {
		info!("Saved chain to {:?}", to);
	    }
	}
	if state.has_shutdown() {
	    break;
	}
	OptionFuture::from(interval.map(|interval| time::delay_for(interval))).await;
	state.when().notified().await;
    }
    trace!("Saver exiting");
}

/// Try to load a chain from this path
pub async fn load(from: impl AsRef<Path>) -> io::Result<Chain<String>>
{
    debug!("Loading chain from {:?}", from.as_ref());
    let file = OpenOptions::new()
        .read(true)
        .open(from).await?;
    let mut whole = Vec::new();
    let mut file = AsyncReadDecompressor::new(file)?;
    tokio::io::copy(&mut file, &mut whole).await?;
    whole.flush().await?;
    serde_cbor::from_slice(&whole[..])
	.map_err(|e| io::Error::new(io::ErrorKind::InvalidInput, e))
}
